#lang racket

;;;; 8. cvičenie z predmetu 2-AIN-285 Symbolické programovanie a LISP
;;;;
;;;; Meno a priezvisko:
;;;; Verzia DrRacket: 6.10.1
;;;; Dátum: 23.4.2018
;;;; WWW: http://www.ii.fmph.uniba.sk/cl/view.php/courses/2-AIN-285-lisp/1718ls/?lang=sk 
;;;; Kontakt: Jan.Komara(at)fmph.uniba.sk 
;;;;
;;;; Úvodná poznámka. Toto cvičenie je venované týmto témam:
;;;; - metainterpreter programovacieho jazyka LISP.
;;;; - imperatívne programovanie.
;;;; Zadanie sa skladá z týchto súborov: 
;;;; - Súbory ex*.rkt obsahujú úlohy, ktoré máte riešiť na tomto cvičení.
;;;;
;;;; Hodnotenie.
;;;; - Počet bodov za úlohy v tomto súbore: 10
;;;; - Počet bodov za prémiove úlohy v tomto súbore: 0
;;;; - Počet bodov za projektové úlohy v tomto súbore: 0
;;;; Odovzdať do 29.4.2018 (nedeľa) do 12:00.

;; [Lisp]

(require racket/undefined)
(define ... undefined)
(define (query f m n) (for/list ([i (in-range m (+ n 1))]) (list i (f i))))

;; Literatúra.
;; [1] Hal Abelson and Jerry Sussman and Julie Sussman.
;;     Structure and Interpretation of Computer Programs.
;;     MIT Press, second edition, 1996. 

;; (define nil null)
;; (define apply-in-underlying-scheme apply)
(require scheme/mpair)
;(require compatibility/mlist)

;;; Kapitola. Metainterpreter programovacieho jazyka LISP.

;;; Sekcia. Vyhodnocovanie výrazov v prostredí väzieb.

;;; Podsekcia. Aritmetizácia.

;; Rozlišovacie položky zoznamových konštruktorov.

(define (tagged-list? exp tag)
  (and (pair? exp)
       (equal? (car exp) tag)))

;; Aritmetizácia výrazov (schéma). 

(define (make-<variant> ...) (list '<variant> ...))     ; konštruktor
(define (<variant>? exp) (tagged-list? exp '<variant>)) ; predikát
(define (<variant>-<selector> exp) ...)                 ; selektor(y)
(define (eval-<variant> exp env) ...)                   ; vyhodnocovanie

;;; Podsekcia. Vyhodnocovanie výrazov v prostredí väzieb.

;; Pokyny. Pozri tiež predchadzajúce cvičenie.

;; Operácia my-eval(exp,env) vyhodnotí výraz exp v prostredí väzieb env.

(define (my-eval exp env)
  (cond [(<variant>? exp) (eval-<variant> exp env)]
        [else (error "Unknown expression type -- EVAL" exp)]))

;;; Podsekcia. Operácie nad prostredím väzieb.

;; Základné operácie.

(define the-empty-environment '())
(define (empty-environment? env) (null? env))
(define (enclosing-environment env) (mcdr env))
(define (first-frame env) (mcar env))
(define the-empty-frame '())
(define (binding-value binding) (mcdr binding))

;; Modifikácia prostredia väzieb.

(define (set-first-frame! env frame) (set-mcar! env frame))
(define (set-binding-value! binding val) (set-mcdr! binding val))

;; Rozšírenie prostredia väzieb.

(define (make-environment frame env)
  (mcons frame env))

(define (add-to-frame-let binding frame)
  (mcons (mcons (car binding) (cadr binding)) frame))

(define (add-to-frame-apply var val frame)
  (mcons (mcons var val) frame))

(define (extend-environment-let bindings env)
  (define (loop bindings frame)
    (if (null? bindings)
        (make-environment frame env)
        (loop (cdr bindings) (add-to-frame-let (car bindings) frame))))
  (loop bindings the-empty-frame))

(define (extend-environment-apply vars vals env)
  (define (loop vars vals frame)
    (if (and (null? vars) (null? vals))
        (make-environment frame env)
        (loop (cdr vars) (cdr vals) (add-to-frame-apply (car vars) (car vals) frame))))
  (loop vars vals the-empty-frame))

;; Globálne prostredie väzieb.

(define (make-primitive-procedure proc)
  (list (string->symbol "primitive procedure") proc))

(define frame-of-primitive-procedures
  (mlist
   (mcons 'zero? (make-primitive-procedure zero?))
   (mcons '+ (make-primitive-procedure +))
   (mcons '- (make-primitive-procedure -))
   (mcons '* (make-primitive-procedure *))
   (mcons 'quotient (make-primitive-procedure quotient))
   (mcons 'remainder (make-primitive-procedure remainder))
   (mcons '= (make-primitive-procedure =))
   (mcons '<= (make-primitive-procedure <=))
   (mcons '< (make-primitive-procedure <))
   (mcons '>= (make-primitive-procedure >=))
   (mcons '> (make-primitive-procedure >))
   (mcons '... (make-primitive-procedure undefined))
   (mcons '#t (make-primitive-procedure #t))
   (mcons '#f (make-primitive-procedure #f))
   ))

(define the-global-environment
  (make-environment frame-of-primitive-procedures the-empty-environment))

;;; Podsekcia. Samo vyhodnocujúce sa výrazy.

;; Predikát.

(define (self-evaluating? exp) ...)

;; Vyhodnocovanie.

(define (eval-self-evaluating exp env) ...)

;; Testovanie.

;; (my-eval 2016 the-global-environment)
;; (my-eval "qwerty" the-global-environment)
;; (my-eval #t the-global-environment)

;;; Podsekcia. Premenné.

;; Predikát.

(define (variable? exp) ...)

;; Vyhodnocovanie.

(define (lookup-variable-value var env) ...)

(define (eval-variable exp env) ...)

;; Testovanie.

;; (my-eval 'x (extend-environment-let '([x 1]) the-global-environment))
;; (my-eval 'x (extend-environment-let '([x 1] [y 2]) the-global-environment))
;; (my-eval 'x (extend-environment-let '([y 2] [x 1]) the-global-environment))
;; (my-eval 'x (extend-environment-let '([x 1]) (extend-environment-let '([x 2]) the-global-environment)))

;;; Podsekcia. Špeciálna forma if (podmienkové výrazy).

;; Pokyny. Pozri predchadzajúce cvičenie.

;; Konštruktor.

(define (make-if predicate consequent alternative)
  (list 'if predicate consequent alternative))

;; Predikát.

(define (if? exp) ...)

;; Selektor(y).

(define (if-predicate exp) ...)
(define (if-consequent exp) ...)
(define (if-alternative exp) ...)

;; Vyhodnocovanie.

(define (eval-if exp env) ...)

;; Testovanie.

;; (my-eval '(if x y z) (extend-environment-let '([x #t] [y 1] [z 2]) the-global-environment))
;; (my-eval '(if x y z) (extend-environment-let '([x #f] [y 2] [z 1]) the-global-environment))

;;; Podsekcia. Špeciálna forma let (priradenie do lokálnej premennej).

;; Konštruktor.

(define (make-let bindings body) (list 'let bindings body))

;; Predikát.

(define (let? exp) ...)

;; Selektor(y).

(define (let-bindings exp) ...)
(define (let-body exp) ...)

;; Vyhodnocovanie.

(define (eval-let-bindings bindings env) ...)

(define (eval-let exp env) ...)

;; Testovanie.

;; (my-eval '(let ([x 1]) x) the-global-environment)
;; (my-eval '(let ([x 1] [y 2]) x) the-global-environment)
;; (my-eval '(let ([y 2] [x 1]) x) the-global-environment)
;; (my-eval '(let ([x 2]) (let ([x 1]) x)) the-global-environment)

;;; Podsekcia. Špeciálna forma define (definície).

;; Konštruktor.
;; Syntax: <head> = <var> | (<var> <parameter₁> ... <parameterₙ>)

(define (make-definition head body) (list 'define head body))

;; Predikát.

(define (definition? exp) ...)

;; Selektor(y).

(define (definition-variable exp) ...)
(define (definition-value exp) ...)

;; Vyhodnocovanie.

(define (define-variable! var val env) ...)
(define (eval-definition exp env) ...)

;;; Podsekcia. Špeciálna forma lambda (operátor abstrakcie).

;; Konštruktor.

(define (make-lambda parameters body) (list 'lambda parameters body))

;; Predikát.

(define (lambda? exp) ...)

;; Selektor(y).

(define (lambda-parameters exp) ...)
(define (lambda-body exp) ...)

;; Vyhodnocovanie.

(define (eval-lambda exp env) ...)

;;; Podsekcia. Aplikatívne výrazy.

;; Konštruktor.

(define (make-application operator operands) (cons operator operands))

;; Predikát.

(define (application? exp) (pair? exp))

;; Selektor(y).

(define (operator exp) (car exp))
(define (operands exp) (cdr exp))

;; Vyhodnocovanie.

(define (eval-application exp env) ...)

;;; Podsekcia. Programy.

;; Konštruktor.

;; (define (make-primitive-procedure proc) (list (string->symbol "primitive procedure") proc))
(define (make-compound-procedure parameters body env) (list (string->symbol "compound procedure") parameters body env))

;; Predikát.

(define (primitive-procedure? proc) ...)

(define (compound-procedure? proc) ...)

;; Selektor(y).

(define (primitive-procedure-implementation proc) ...)

(define (compound-procedure-parameters proc) ...)
(define (compound-procedure-body proc) ...)
(define (compound-procedure-environment proc) ...)

;; Vyhodnocovanie.

(define (apply-primitive-procedure proc args) ...)
(define (apply-compound-procedure proc args) ...)
(define (my-apply proc args) ...)

;; Podsekcia. Testovanie.

;; (my-eval '(+ 1 2) the-global-environment)
;; (my-eval '(let ([x 1]) (+ x 2)) the-global-environment)
;; (my-eval '(let ([x 1] [y 2]) (+ x y)) the-global-environment)
;; (my-eval '(let ([x 2]) (let ([x 3]) x)) the-global-environment)
;; (my-eval '(let ([x (+ 1 2)]) x) the-global-environment)

;;; Podsekcia. "Read–eval–print loop."

(define (my-read)
  (display "input : ")
  (read))

(define (my-print exp)
  (display "output: ")
  (if (compound-procedure? exp)
      (display '\#<procedure>)
      (display exp))
  (newline))

(define (driver-loop)
  (my-print (my-eval (my-read) the-global-environment))
  (driver-loop))

;; Ak chceme používať REPL, tak nasledujúci riadok treba odkomentovať.
;; (driver-loop)

;;; Podsekcia. Testovanie.

;; (my-eval '(+ 1 2) the-global-environment)
;; (my-eval '(let ([x 1]) (+ x 2)) the-global-environment)
;; (my-eval '(let ([x 1] [y 2]) (+ x y)) the-global-environment)
;; (my-eval '(let ([x (+ 1 2)]) x) the-global-environment)

;;; Podsekcia. Testovanie.

;; Táto časť slúži výlučne na testovanie. Nehodnotí sa!

;; Úloha. Zostrojte program, ktorý pre dané číslo vráti jeho druhú mocninu.

;; (my-eval '... the-global-environment)

;(my-eval '(define (square x) (* x x)) the-global-environment)

;; (my-eval '(square 5) the-global-environment)

;; Úloha. Zostrojte program, ktorý pre dané dve čísla vypočíta súčet ich
;; štvorcov. 

;; (my-eval '... the-global-environment)

;(my-eval '(define (sum-of-squares x y) (+ (square x) (square y))) the-global-environment)

;; (my-eval '(sum-of-squares 3 4) the-global-environment)

;; Úloha. Zostrojte program, ktorý pre dané dve čísla vráti ich minimum.

;; (my-eval '... the-global-environment)

;(my-eval '(define (min x y) (if (<= x y) x y)) the-global-environment)

;; (my-eval '(min 1 2) the-global-environment)

;; Úloha. Definujte primitívnou rekurziou sumačnú funkciu.

;; (my-eval '... the-global-environment)

;(my-eval '(define (sum n) (if (= n 0) 0 (+ (sum (- n 1)) n))) the-global-environment)

;; (my-eval '(sum 100) the-global-environment)

;; Úloha. Definujte primitívnou rekurziou funkciu faktoriál.

;; (my-eval '... the-global-environment)

;(my-eval '(define (fact n) (if (= n 0) 1 (* n (fact (- n 1))))) the-global-environment)

;; (my-eval '(fact 6) the-global-environment)

;; Úloha. Implementujte Euklidov algoritmus pre výpočet najväčšieho spoločného
;; deliteľa dvoch prirodzených čísel.

;; (my-eval '... the-global-environment)

;(my-eval '(define (gcd x y) (if (= x 0) y (gcd (remainder y x) x))) the-global-environment)

;; (my-eval '(gcd 21 12) the-global-environment)

;; Úloha. Nájdite iteratívny algoritmus pre výpočet Fibonacciho postupnosti.

;; (my-eval '... the-global-environment)

;(my-eval '(define (fib-iter n a b) (if (= n 0) a (fib-iter (- n 1) (+ a b) a))) the-global-environment)
;(my-eval '(define (fib n) (if (= n 0) 0 (fib-iter (- n 1) 1 0))) the-global-environment)

;;(my-eval '(fib 10) the-global-environment)
